package com.elitesland.system.service.impl;

import com.elitesland.core.base.ApiCode;
import com.elitesland.core.base.BaseModel;
import com.elitesland.core.base.PagingVO;
import com.elitesland.core.exception.BusinessException;
import com.elitesland.system.convert.SysRoleConvert;
import com.elitesland.system.convert.SysUserConvert;
import com.elitesland.system.entity.QSysRoleDO;
import com.elitesland.system.entity.SysRoleDO;
import com.elitesland.system.param.SysRoleNewParam;
import com.elitesland.system.param.SysRoleQueryParam;
import com.elitesland.system.param.SysRoleUpdateParam;
import com.elitesland.system.repo.SysRoleRepo;
import com.elitesland.system.repo.SysRoleRepoProc;
import com.elitesland.system.service.SysPermissionService;
import com.elitesland.system.service.SysRoleService;
import com.elitesland.system.vo.SysRoleVO;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import lombok.RequiredArgsConstructor;
import lombok.val;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * <pre>
 * [功能说明]
 * </pre>
 *
 * @author Mir
 * @date 2020/7/5
 */
@Service
@RequiredArgsConstructor
public class SysRoleServiceImpl implements SysRoleService {

    private final SysRoleRepo sysRoleRepo;

    private final SysPermissionService sysPermissionService;

    @Override
    public PagingVO<SysRoleVO> search(SysRoleQueryParam param) {
        val roleItor = sysRoleRepo.findAll(SysRoleRepoProc.where(param), param.getPageRequest());

        val roles = roleItor.getContent();

        List<SysRoleVO> roleVos = roles.stream().map(r -> {
            val vo = new SysRoleVO();
            vo.setId(r.getId());
            vo.setCode(r.getCode());
            vo.setName(r.getName());
            vo.setEnabled(r.getEnabled());
            vo.setUsers(r.getUsers().stream().map(SysUserConvert.INSTANCE::doToVo).collect(Collectors.toList()));
            vo.setPermIds(r.getPermissions().stream().map(BaseModel::getId).collect(Collectors.toList()));

            return vo;
        }).collect(Collectors.toList());
        return PagingVO.<SysRoleVO>builder()
                .total(roleItor.getTotalElements())
                .records(roleVos)
                .build();
    }

    @Override
    public Optional<SysRoleDO> one(Long roleId) {
        return sysRoleRepo.findById(roleId);
    }

    @Override
    @Transactional
    public Long create(SysRoleNewParam role) {
        val sysRole = QSysRoleDO.sysRoleDO;

        val existRolesItor = sysRoleRepo.findAll(sysRole.code.in(role.getCode()));
        val existRoles = Lists.newArrayList(existRolesItor);
        if (existRoles.size() > 0) {
            val codes = existRoles.stream().map(SysRoleDO::getCode).collect(Collectors.joining(", "));
            throw new BusinessException("角色代码重复: " + codes);
        }
        val permissions = sysPermissionService.listAllPermissionsByIds(role.getPermIds());

        val tempRole = SysRoleConvert.INSTANCE.newParamToDO(role);
        tempRole.setPermissions(permissions);

        val newRole = sysRoleRepo.save(tempRole);

        return newRole.getId();
    }

    @Override
    @Transactional(rollbackFor = {Exception.class})
    public void update(SysRoleUpdateParam reqRole) {
        val sysRole = QSysRoleDO.sysRoleDO;

        val roleOpt = sysRoleRepo.findById(reqRole.getId());
        if (roleOpt.isEmpty()) {
            throw new BusinessException("待更新的角色不存在");
        }

        val roleCount = sysRoleRepo.count(sysRole.code.eq(reqRole.getCode()).and(sysRole.id.ne(reqRole.getId())));
        if (roleCount > 0) {
            throw new BusinessException("角色编码：" + reqRole.getCode() + "， 已经存在");
        }

        val role = roleOpt.get();
        role.setCode(reqRole.getCode());
        role.setName(reqRole.getName());
        role.setEnabled(reqRole.getEnabled());

        val permissions = sysPermissionService.listAllPermissionsByIds(reqRole.getPermIds());
        role.setPermissions(permissions);

        sysRoleRepo.save(role);
    }

    @Override
    public Set<SysRoleDO> listByRoleIds(List<Long> roleIds) {
        val role = QSysRoleDO.sysRoleDO;
        val roleItor = sysRoleRepo.findAll(role.id.in(roleIds));
        return Sets.newHashSet(roleItor);
    }

    @Override
    public List<SysRoleDO> listAll() {
        return sysRoleRepo.findAll();
    }

    @Override
    public List<String> listMenusById(Long id) {
        return one(id).map(r -> r.getPermissions().stream()
                .filter(p -> p.getPermType() == 0).map(p -> p.getId().toString())
                .collect(Collectors.toList())
        ).orElse(new ArrayList<>());
    }

    @Override
    public List<String> listActionsById(Long id) {
        return one(id).map(r -> r.getPermissions().stream()
                .filter(p -> p.getPermType() == 1).map(p -> p.getId().toString())
                .collect(Collectors.toList())
        ).orElse(new ArrayList<>());
    }

    @Override
    public Boolean switchRoleStatus(Long id) {
        val roleOpt = sysRoleRepo.findById(id);

        return roleOpt.map(r -> {
            r.setEnabled(!r.getEnabled());
            sysRoleRepo.save(r);
            return r.getEnabled();
        }).orElseThrow(
                new BusinessException(ApiCode.FAIL, "ID为：" + id + "，的角色不存在")
        );
    }
}
