package com.elitesland.system.service.impl;

import com.elitesland.core.base.ApiCode;
import com.elitesland.core.base.BaseModel;
import com.elitesland.core.base.PagingVO;
import com.elitesland.core.exception.BusinessException;
import com.elitesland.security.service.entity.JwtUserDto;
import com.elitesland.system.convert.SysPermissionConvert;
import com.elitesland.system.convert.SysUserConvert;
import com.elitesland.system.entity.QSysUserDO;
import com.elitesland.system.entity.SysPermissionDO;
import com.elitesland.system.entity.SysRoleDO;
import com.elitesland.system.entity.SysUserDO;
import com.elitesland.system.param.SysUserNewParam;
import com.elitesland.system.param.SysUserQueryParam;
import com.elitesland.system.param.SysUserUpdateParam;
import com.elitesland.system.repo.SysUserRepo;
import com.elitesland.system.repo.SysUserRepoProc;
import com.elitesland.system.service.SysPermissionService;
import com.elitesland.system.service.SysRoleService;
import com.elitesland.system.service.SysUserService;
import com.elitesland.system.vo.SysPermissionVO;
import com.elitesland.system.vo.SysUserVO;
import lombok.RequiredArgsConstructor;
import lombok.val;
import org.apache.commons.lang3.StringUtils;
import org.springframework.security.core.authority.AuthorityUtils;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.User;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.*;
import java.util.stream.Collectors;

/**
 * <pre>
 * [功能说明]
 * </pre>
 *
 * @author Mir
 * @date 2020/6/16
 */
@Service
@RequiredArgsConstructor
public class SysUserServiceImpl implements SysUserService {

    private final SysUserRepo sysUserRepo;

    private final PasswordEncoder passwordEncoder;

    private final SysRoleService sysRoleService;

    private final SysPermissionService sysPermissionService;

    private final SysUserRepoProc sysUserRepoProc;

    @Override
    public PagingVO<SysUserVO> search(SysUserQueryParam param) {

        val users = sysUserRepo.findAll(sysUserRepoProc.where(param), param.getPageRequest());

        val userVos = users.getContent().stream().map(u -> {
            SysUserVO userVo = SysUserConvert.INSTANCE.doToVo(u);
            userVo.setRoleNames(u.getRoles().stream().map(SysRoleDO::getName).collect(Collectors.joining(",")));
            return userVo;
        }).collect(Collectors.toList());

        return PagingVO.<SysUserVO>builder()
                .total(users.getTotalElements())
                .records(userVos)
                .build();
    }

    @Override
    public Optional<SysUserDO> getById(Long id) {
        return sysUserRepo.findById(id);
    }

    @Override
    public Optional<SysUserDO> getUserByUsernameSec(String username) {
        val sysUser = QSysUserDO.sysUserDO;
        val predicate = sysUser.username.eq(username);
        return sysUserRepo.findOne(predicate);
    }

    @Override
    public Optional<SysUserVO> getUserByUsername(String username) {
        val userOpt = getUserByUsernameSec(username);

        return userOpt.map(SysUserConvert.INSTANCE::doToVo);
    }

    @Override
    public boolean isUserExists(String username) {
        val userOpt = getUserByUsernameSec(username);
        return userOpt.isPresent();
    }

    @Override
    public Optional<UserDetails> loadUserDetailsByUsername(String username) {
        val userOpt = getUserByUsernameSec(username);
        return userOpt.map(o -> User.builder()
                .username(o.getUsername())
                .password(o.getPassword())
                .authorities(AuthorityUtils.createAuthorityList(o.getRoles().stream().map(SysRoleDO::getCode).toArray(String[]::new)))
                .build());
    }

    @Override
    @Transactional
    public Long create(SysUserNewParam user) throws BusinessException {
        val oldUserOpt = getUserByUsernameSec(user.getUsername());
        if (oldUserOpt.isPresent()) {
            throw new BusinessException("用户：" + user.getUsername() + ", 已经存在");
        }

        if (StringUtils.isNotBlank(user.getMobile())) {
            val mobileOpt = sysUserRepo.findByMobile(user.getMobile());
            if (mobileOpt.isPresent()) {
                throw new BusinessException("相同的手机号码已经存在");
            }
        }

        if (StringUtils.isNotBlank(user.getEmail())) {
            val emailOpt = sysUserRepo.findByEmail(user.getEmail());
            if (emailOpt.isPresent()) {
                throw new BusinessException("相同的邮箱地址已经存在");
            }
        }

        val sysUser = SysUserConvert.INSTANCE.newToDo(user);
        sysUser.setPassword(passwordEncoder.encode(user.getPassword()));

        val roles = sysRoleService.listByRoleIds(user.getRoleIds());
        sysUser.setRoles(roles);
        val newUser = sysUserRepo.save(sysUser);
        return newUser.getId();
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void update(SysUserUpdateParam user) {
        val userOpt = sysUserRepo.findById(user.getId());
        if (userOpt.isEmpty()) {
            throw new BusinessException("ID为：" + user.getId() + "， 的用户不存在");
        }
        val u = userOpt.get();

        if (StringUtils.isNotBlank(user.getFirstName())) {
            u.setFirstName(user.getFirstName());
        }
        if (StringUtils.isNotBlank(user.getLastName())) {
            u.setLastName(user.getLastName());
        }
        u.setEnabled(user.getEnabled());

        val sysUser = QSysUserDO.sysUserDO;
        if (StringUtils.isNotBlank(user.getMobile())) {
            val mobileCount = sysUserRepo.count(sysUser.mobile.eq(user.getMobile()).and(sysUser.id.ne(user.getId())));
            if (mobileCount > 0) {
                throw new BusinessException("手机号码：" + user.getMobile() + ", 已经存在");
            }
            u.setMobile(user.getMobile());
        }
        if (StringUtils.isNotBlank(user.getEmail())) {
            val emailCount = sysUserRepo.count(sysUser.email.eq(user.getEmail()).and(sysUser.id.ne(user.getId())));
            if (emailCount > 0) {
                throw new BusinessException("电子邮箱：" + user.getEmail() + ", 已经存在");
            }
            u.setEmail(user.getEmail());
        }

        if (StringUtils.isNotBlank(user.getPassword())) {
            u.setPassword(passwordEncoder.encode(user.getPassword()));
        }

        val roles = sysRoleService.listByRoleIds(user.getRoleIds());
        u.setRoles(roles);

        sysUserRepo.save(u);
    }

    @Override
    public void changePasswordNoSec(Long userId, String newPassword) {
        sysUserRepo.findById(userId).ifPresent(u -> {
            u.setPassword(passwordEncoder.encode(newPassword));
            sysUserRepo.save(u);
        });
    }

    @Override
    @Transactional(rollbackFor = {Exception.class})
    public Boolean switchUserStatus(Long id) {
        val userOpt = sysUserRepo.findById(id);
        if (userOpt.isEmpty()) {
            throw new BusinessException(ApiCode.FAIL, "ID为：" + id + "，的用户不存在");
        }
        val user = userOpt.get();
        user.setEnabled(!user.getEnabled());
        sysUserRepo.save(user);
        return user.getEnabled();
    }

    @Override
    public Set<SysRoleDO> listRolesByUsername(String username) {
        val userOpt = getUserByUsernameSec(username);
        return userOpt.map(SysUserDO::getRoles).orElse(new HashSet<>());
    }

    @Override
    public Optional<SysUserDO> current() {
        val principal = SecurityContextHolder.getContext().getAuthentication().getPrincipal();
        if (principal instanceof JwtUserDto) {
            val jwtUserDto = (JwtUserDto) principal;
            val user = jwtUserDto.getUser();
            user.setPassword("");
            return Optional.of(user);
        }
        return Optional.empty();
    }

    @Override
    public List<SysPermissionVO> currentMenu() {
        return current().map(u -> {
            val permSet = u.getRoles().stream().flatMap(r -> r.getPermissions().stream()
                    .filter(p -> p.getPermType() == 0))
                    .collect(Collectors.toSet());

            val idSet = permSet.stream().map(BaseModel::getId).collect(Collectors.toSet());
            val pidSet = permSet.stream().flatMap(p -> Arrays.stream(p.getPath().split("/")).filter(StringUtils::isNotBlank))
                    .map(Long::parseLong).collect(Collectors.toSet());
            idSet.addAll(pidSet);

            val permissions = sysPermissionService.listAllPermissionsByIds(new ArrayList<>(idSet));

            val sortPerms = new ArrayList<SysPermissionDO>();
            permissions.stream().filter(p -> p.getParentId() == 0L)
                    .sorted(Comparator.comparing(SysPermissionDO::getSortNo)).forEach(tp -> {
                sortPerms.add(tp);

                permissions.stream().filter(p -> p.getParentId().equals(tp.getId()))
                        .sorted(Comparator.comparing(SysPermissionDO::getSortNo)).forEach(sp -> {
                    sortPerms.add(sp);

                    val thirdPerms = permissions.stream().filter(p -> p.getParentId().equals(sp.getId()))
                            .sorted(Comparator.comparing(SysPermissionDO::getSortNo))
                            .collect(Collectors.toList());

                    sortPerms.addAll(thirdPerms);
                });
            });

            return sortPerms.stream().map(SysPermissionConvert.INSTANCE::doToVO).collect(Collectors.toList());
        }).orElse(new ArrayList<>());
    }

    @Override
    public List<SysPermissionVO> currentActions() {
        return current().map(u -> {
            val actionSet = u.getRoles().stream().flatMap(r -> r.getPermissions().stream()
                    .filter(p -> p.getPermType() == 1))
                    .collect(Collectors.toSet());

            return actionSet.stream().map(SysPermissionConvert.INSTANCE::doToVO).collect(Collectors.toList());
        }).orElse(new ArrayList<>());
    }

    @Override
    public List<SysPermissionVO> currentMenuActions(Long menuId) {
        return current().map(u -> {
            val actionSet = u.getRoles().stream().flatMap(r -> r.getPermissions().stream())
                    .filter(p -> p.getParentId().equals(menuId)).collect(Collectors.toSet());

            return actionSet.stream().map(SysPermissionConvert.INSTANCE::doToVO).collect(Collectors.toList());
        }).orElse(new ArrayList<>());
    }
}
